package article

import (
	"bytes"
	"fmt"
	"io"
	"io/ioutil"
	"regexp"
	"sync"
	"testing"
)

var reHelloWorld = regexp.MustCompile(`Hello, ([^!]+)!`)

func FastMatchHelloWorld(s string) bool {
	return reHelloWorld.MatchString(s)
}

func SlowMatchHelloWorld(s string) bool {
	// このままでは
	var re = regexp.MustCompile(`Hello, ([^!]+)!`)
	return re.MatchString(s)
}

func BenchmarkMatchHelloWorld(b *testing.B) {
	data := map[string]func(string) bool{
		"Slow": SlowMatchHelloWorld,
		"Fast": FastMatchHelloWorld,
	}
	for name, f := range data {
		b.Run(name, func(b *testing.B) {
			for i := 0; i < b.N; i++ {
				f(`Hello, World!`)
			}
		})
	}
}

func BenchmarkAppend(b *testing.B) {
	var max = 255
	b.Run("Pre-Allocate", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			buf := make([]byte, 0, max)
			for j := 0; j < max; j++ {
				buf = append(buf, byte(j))
			}
		}
	})
	b.Run("Allocate-As-You-Go", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			var buf []byte
			for j := 0; j < max; j++ {
				buf = append(buf, byte(j))
			}
		}
	})
}

func ByValue(buf bytes.Buffer) {
	io.Copy(ioutil.Discard, &buf)
}

func ByReference(buf *bytes.Buffer) {
	io.Copy(ioutil.Discard, buf)
}

func BenchmarkValueVSReference(b *testing.B) {
	var buf bytes.Buffer
	buf.WriteString(`Hello, World!`)
	b.Run("By Value", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			ByValue(buf)
		}
	})
	b.Run("By Reference", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			ByReference(&buf)
		}
	})
}

func MakeByteSlicePool(s int) *sync.Pool {
	return &sync.Pool{
		New: func() interface{} {
			return make([]byte, 0, s)
		},
	}
}

func MakeByteSlice(size int) {
	buf := make([]byte, 0, size)
	ioutil.Discard.Write(buf)
}

func ReuseByteSlice(pool *sync.Pool) {
	buf := pool.Get().([]byte)
	ioutil.Discard.Write(buf)
	pool.Put(buf)
}

func BenchmarkPool(b *testing.B) {
	const (
		smallBufsize = 64
		bigBufsize   = 1 << 16
	)
	var smallPool = MakeByteSlicePool(smallBufsize)
	var bigPool = MakeByteSlicePool(bigBufsize)

	for _, parallel := range []bool{false, true} {
		b.Run(fmt.Sprintf("parallel = %t", parallel), func(b *testing.B) {
			var runner func(*testing.B, func(b *testing.B))
			if parallel {
				runner = func(b *testing.B, f func(b *testing.B)) {
					b.RunParallel(func(pb *testing.PB) {
						for pb.Next() {
							f(b)
						}
					})
				}
			} else {
				runner = func(b *testing.B, f func(b *testing.B)) {
					for i := 0; i < b.N; i++ {
						f(b)
					}
				}
			}

			b.Run("allocate small", func(b *testing.B) {
				runner(b, func(b *testing.B) {
					MakeByteSlice(smallBufsize)
				})
			})
			b.Run("allocate small (via sync.Pool)", func(b *testing.B) {
				runner(b, func(b *testing.B) {
					ReuseByteSlice(smallPool)
				})
			})
			b.Run("allocate big", func(b *testing.B) {
				runner(b, func(b *testing.B) {
					MakeByteSlice(bigBufsize)
				})
			})
			b.Run("allocate big (via sync.Pool)", func(b *testing.B) {
				runner(b, func(b *testing.B) {
					ReuseByteSlice(bigPool)
				})
			})
		})
	}
}
